package net.sys42.dakedres.grafting.blocks;

import eu.pb4.polymer.resourcepack.api.PolymerModelData;
import eu.pb4.sgui.api.ClickType;
import eu.pb4.sgui.api.elements.GuiElement;
import eu.pb4.sgui.api.elements.GuiElementInterface;
import eu.pb4.sgui.api.gui.SimpleGui;
import eu.pb4.sgui.virtual.inventory.VirtualSlot;
import net.minecraft.block.BlockState;
import net.minecraft.block.CrafterBlock;
import net.minecraft.block.entity.LockableContainerBlockEntity;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.entity.player.PlayerInventory;
import net.minecraft.inventory.CraftingResultInventory;
import net.minecraft.inventory.Inventories;
import net.minecraft.inventory.Inventory;
import net.minecraft.inventory.RecipeInputInventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NbtCompound;
import net.minecraft.recipe.RecipeMatcher;
import net.minecraft.recipe.input.CraftingRecipeInput;
import net.minecraft.registry.RegistryWrapper;
import net.minecraft.screen.NamedScreenHandlerFactory;
import net.minecraft.screen.ScreenHandler;
import net.minecraft.screen.ScreenHandlerType;
import net.minecraft.screen.slot.CraftingResultSlot;
import net.minecraft.screen.slot.Slot;
import net.minecraft.screen.slot.SlotActionType;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.text.Text;
import net.minecraft.util.collection.DefaultedList;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.sys42.dakedres.grafting.Grafting;
import net.sys42.dakedres.grafting.PlacementMatrixUtils;
import net.sys42.dakedres.grafting.ui.Icons;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;

public class WorkbenchBlockEntity extends LockableContainerBlockEntity implements RecipeInputInventory {

	public static final String NBT_GRID_ROTATION_KEY = "GridRotation";
	public static final int GRID_SIZE = 9;
	public static final int GRID_WIDTH = 3;

	private DefaultedList<ItemStack> grid = DefaultedList.ofSize(GRID_SIZE, ItemStack.EMPTY);
	private int gridRotation = 0;
	private final ArrayList<Gui> guis = new ArrayList<>();
	public WorkbenchBlockEntity(BlockPos blockPos, BlockState blockState) {
		super(Grafting.WORKBENCH_BLOCK_ENTITY_TYPE, blockPos, blockState);
	}

	public void openMenu(ServerPlayerEntity playerEntity) {
		playerEntity.openHandledScreen(new NamedScreenHandlerFactory() {
			@Override
			public Text getDisplayName() {
				return WorkbenchBlockEntity.this.getDisplayName();
			}

			@Override
			public @Nullable ScreenHandler createMenu(int syncId, PlayerInventory playerInventory, PlayerEntity player) {
				return WorkbenchBlockEntity.this.createScreenHandler(syncId, playerInventory);
			}
		});
	}

	@Override
	protected void readNbt(NbtCompound nbt, RegistryWrapper.WrapperLookup registryLookup) {
		super.readNbt(nbt, registryLookup);
		this.gridRotation = nbt.getByte(NBT_GRID_ROTATION_KEY);
		this.grid = DefaultedList.ofSize(this.size(), ItemStack.EMPTY);
	}

	@Override
	protected void writeNbt(NbtCompound nbt, RegistryWrapper.WrapperLookup registryLookup) {
		super.writeNbt(nbt, registryLookup);
		nbt.putByte(NBT_GRID_ROTATION_KEY, (byte) gridRotation);
		Inventories.writeNbt(nbt, this.grid, registryLookup);
	}

	@Override
	protected Text getContainerName() {
		// TODO: localize
		return Text.of("Workbench");
	}

	@Override
	public int getWidth() {
		return GRID_WIDTH;
	}

	@Override
	public int getHeight() {
		return GRID_WIDTH;
	}

	@Override
	public DefaultedList<ItemStack> getHeldStacks() {
		return this.grid;
	}

	@Override
	protected void setHeldStacks(DefaultedList<ItemStack> inventory) {
		this.grid = inventory;
	}

	@Override
	protected ScreenHandler createScreenHandler(int syncId, PlayerInventory playerInventory) {
		return createScreenHandler(syncId, playerInventory, playerInventory.player);
	}

	protected ScreenHandler createScreenHandler(int syncId, PlayerInventory playerInventory, PlayerEntity player) {
		return new Gui((ServerPlayerEntity) player).openAsScreenHandler(syncId, playerInventory, player);
	}

	@Override
	public int size() {
		return GRID_SIZE;
	}

	public boolean hasEmptySlots() {
		for(ItemStack itemStack : grid) {
			if (itemStack.isEmpty()) {
				return true;
			}
		}

		return false;
	}

	private int[] getPlacementMatrix() {
		return PlacementMatrixUtils.rotate(PlacementMatrixUtils.getPlacements(), gridRotation);
	}

	public void placeStack(ItemStack stack) {
		int[] placements = getPlacementMatrix();

		for(int i = 0; i < GRID_SIZE; i++) {
			int p = placements[i];
			if(getStack(p).isEmpty()) {
				setStack(p, stack);
				distributeStack(stack, 0);
				return;
			}
		}
	}

	public void distributeStack(ItemStack stack) {
		distributeStack(stack, stack.getCount());
	}

	public void distributeStack(ItemStack stack, int count) {
		ArrayList<ItemStack> candidates = new ArrayList<>();
		int totalCount = count;
		int maxTotalCount = 0;
		int[] placements = getPlacementMatrix();

		for(int i = 0; i < GRID_SIZE; i++) {
			int p = placements[i];
			ItemStack ps = getStack(p);
			if(!ps.isEmpty() && ItemStack.areItemsAndComponentsEqual(ps, stack)) {
				totalCount += ps.getCount();
				maxTotalCount += ps.getMaxCount();
				ps.setCount(0);
				candidates.add(ps);
			}
		}

		int remainder = 0;
		if(totalCount > maxTotalCount) {
			remainder = totalCount - maxTotalCount;
			totalCount = maxTotalCount;
		}
		int i = 0;
		while(totalCount > 0) {
			candidates.get(i % candidates.size()).increment(1);
			totalCount--;
			i++;
		}

		stack.decrement(count - remainder);
		markDirty();
	}

	protected int lastPriorityIndexAlike(ItemStack stack) {
		int[] placements = getPlacementMatrix();

		for(int i = placements.length - 1; i >= 0; i--) {
			int p = placements[i];
			ItemStack ps = getStack(p);
			if(ItemStack.areItemsAndComponentsEqual(ps, stack)) {
				return p;
			}
		}

		return -1;
	}

	public ItemStack removeStack(ItemStack stack, int amount) {
		int index = lastPriorityIndexAlike(stack);

		if(index >= 0) {
			ItemStack o = removeStack(index, amount);
			distributeStack(o, 0);
			return o;
		}

		return ItemStack.EMPTY;
	}

	public boolean hasItemsLike(ItemStack stack) {
		for(ItemStack itemStack : grid) {
			if(ItemStack.areItemsAndComponentsEqual(itemStack, stack)) {
				return true;
			}
		}

		return false;
	}

//		Based on CrafterScreenHandler#updateResult
	public ItemStack getResult(ServerPlayerEntity player) {
		World world = player.getWorld();
		CraftingRecipeInput craftingRecipeInput = createRecipeInput();
		return CrafterBlock.getCraftingRecipe(world, craftingRecipeInput)
				.map(recipe -> recipe.value()
						.craft(craftingRecipeInput, world.getRegistryManager())
				)
				.orElse(ItemStack.EMPTY);
	}

	@Override
	public void markDirty() {
		for(Gui gui : guis) {
			gui.updateResult();
		}
		super.markDirty();
	}

	@Override
	public void provideRecipeInputs(RecipeMatcher finder) {

	}

	// TODO: Reimplement. Need to prevent any possible desync between the orientation as manipulated here, and the one in getPlacements
	public void rotate(int direction) {
		gridRotation = PlacementMatrixUtils.clampRotation(gridRotation + direction);
		int[] rotationMatrix = PlacementMatrixUtils.ROTATION_MATRICES[PlacementMatrixUtils.clampRotation(direction)];
		DefaultedList<ItemStack> newGrid = DefaultedList.ofSize(grid.size(), ItemStack.EMPTY);

		for(int i = 0; i < grid.size(); i++) {
			newGrid.set(rotationMatrix[i], grid.get(i));
		}

		setHeldStacks(newGrid);
		markDirty();
	}

	public class Gui extends SimpleGui {
		private static final int BUTTONS_LEFT = 1;
		private static final int GRID_LEFT = BUTTONS_LEFT + 1;
		private static final int RESULT_LEFT = GRID_LEFT + GRID_WIDTH + 2;

		private final CraftingResultInventory resultInventory = new CraftingResultInventory();
		public Gui(ServerPlayerEntity player) {
			super(ScreenHandlerType.GENERIC_9X3, player, false);
			guis.add(this);

			this.setTitle(WorkbenchBlockEntity.this.getDisplayName());
			setSlot(BUTTONS_LEFT, new RotateGridButton(Icons.ROTATE_CLOCKWISE,1));
			setSlot(BUTTONS_LEFT + 9, new RotateGridButton(Icons.ROTATE_COUNTER_CLOCKWISE, -1));
			setSlotRedirect(
					RESULT_LEFT + 9,
					new CraftingResultSlot(player, WorkbenchBlockEntity.this, resultInventory, 0, 0, 0)
			);
			for(int i = 0; i < GRID_SIZE; i++) {
				int row = (i / GRID_WIDTH);
				setSlotRedirect(
						GRID_LEFT + (i % GRID_WIDTH) + (row * 9),
						new GridSlot(i)
				);
			}

			updateResult();
		}

		@Override
		public void close() {
			guis.remove(this);
			super.close();
		}

		protected void updateResult() {
			resultInventory.setStack(0, getResult(player));
		}

		/*
		According to ScreenHandler#internalOnSlotClick, quickMove is continually called
		while it returns a non-empty ItemStack that matches the original slot
		 */
		@Override
		public ItemStack quickMove(int index) {
			ItemStack returnStack = ItemStack.EMPTY;
			Slot slot = this.getSlotRedirectOrPlayer(index);

			if(slot != null && slot.hasStack() && !(slot instanceof VirtualSlot)) {
				ItemStack operatingStack = slot.getStack();
				returnStack = operatingStack.copy();

				if(index < this.getVirtualSize()) {
					int i = -1;

					if(slot instanceof GridSlot) {
						i = lastPriorityIndexAlike(operatingStack);
						operatingStack = i == -1 ? ItemStack.EMPTY : removeStack(i);

						if (!this.insertItem(operatingStack, this.getVirtualSize(), this.getVirtualSize() + 36, true)) {
							return ItemStack.EMPTY;
						}
					}

					if (!this.insertItem(operatingStack, this.getVirtualSize(), this.getVirtualSize() + 36, true)) {
						return ItemStack.EMPTY;
					}

					if(slot.getIndex() == i) {
						return ItemStack.EMPTY;
					}
				} else {
					if(hasItemsLike(operatingStack)) {
						distributeStack(operatingStack);
						return ItemStack.EMPTY;
					} else if(hasEmptySlots()) {
						placeStack(operatingStack);
						slot.setStack(ItemStack.EMPTY);
					} else {
						return ItemStack.EMPTY;
					}
				}

				if (operatingStack.isEmpty()) {
					slot.setStack(ItemStack.EMPTY);
				} else {
					slot.markDirty();
				}
			} else if (slot instanceof VirtualSlot) {
				return slot.getStack();
			}

			return returnStack;
		}

		//		- Inserting single items changes the shape
		//		- Inserting a full stack will attempt to distribute the stack, or insert if that stack isn't in the workbench
		public class GridSlot extends Slot {
			public GridSlot(int index) {
				super(WorkbenchBlockEntity.this, index, index, 0);
			}

			@Override
			public ItemStack insertStack(ItemStack stack, int count) {

				if(stack.isEmpty()) {
					return stack;
				}

				int allowedCount = Math.min(count, stack.getCount());
				if(
						(!hasItemsLike(stack) || count == 1) &&
								WorkbenchBlockEntity.this.hasEmptySlots()
				) {
					placeStack(stack.split(allowedCount));
				} else {
					distributeStack(stack, allowedCount);
				}

				return stack;
			}

			@Override
			public ItemStack takeStack(int amount) {
				return removeStack(this.getStack(), amount);
			}
		}

		public class RotateGridButton extends GuiElement {
			public RotateGridButton(PolymerModelData model, int bias) {
				super(
						Icons.getStack(model),
						new GuiElementInterface.ItemClickCallback() {
							@Override
							public void click(int i, ClickType clickType, SlotActionType slotActionType) {
								switch(clickType) {
									case ClickType.MOUSE_LEFT -> rotate(bias);
									case ClickType.MOUSE_RIGHT -> rotate(bias * -1);
								}
							}
						}
				);
			}
		}
	}
}
